from pathlib import _NormalAccessor

from pathliberty.base import AbstractPath, AbstractPathAccessor

class NormalAccessor(_NormalAccessor, AbstractPathAccessor):
    pass

class LocalPath(AbstractPath):
    __slots__ = ()
    accessor_class = NormalAccessor

    def new(self, path: AbstractPath) -> AbstractPath:
        return path



from pathliberty.base import (
    AbstractPathAccessor,
    AbstractPath
)

from pathliberty.local_path import (
    LocalPath,
)

from pathliberty.remote_path import (
    AbstractRemotePath,
)

from pathliberty.ssh_path import (
    SSHAccessor,
    SSHPath
)



from __future__ import annotations
import abc
from pathlib import PosixPath
from typing import Any, Callable

from pathliberty.base import AbstractPath

class AbstractRemotePath(AbstractPath):
    """ Represents a path that exists on a some remote node """
    __slots__ = ()

    @abc.abstractmethod
    def cwd(self): raise NotImplementedError

    @abc.abstractmethod
    def home(self): raise NotImplementedError

    # @abc.abstractmethod
    # def glob(*args): raise NotImplementedError

    # @abc.abstractmethod
    # def rglob(*args): raise NotImplementedError

    @abc.abstractmethod
    def open(*args): raise NotImplementedError

    @abc.abstractmethod
    def touch(*args): raise NotImplementedError

    @abc.abstractmethod
    def is_equal_to(self, other) -> bool:
        raise NotImplementedError

    def _rich_comparison(self, method: Callable, other: Any) -> bool:
        if isinstance(other, AbstractRemotePath):
            if not self.is_equal_to(other):
                return NotImplemented
        return method(other)

    def __eq__(self, other): return self._rich_comparison(super().__eq__, other)
    def __lt__(self, other): return self._rich_comparison(super().__lt__, other)
    def __le__(self, other): return self._rich_comparison(super().__le__, other)
    def __gt__(self, other): return self._rich_comparison(super().__gt__, other)
    def __ge__(self, other): return self._rich_comparison(super().__ge__, other)
    __hash__ = PosixPath.__hash__



from __future__ import annotations

import abc
from pathlib import PosixPath
from typing import Sequence, TypeVar

_PT = TypeVar('_PT')

class AbstractPathAccessor:
    def __init__(self, path_obj: AbstractPath) -> None:
        pass

    @staticmethod
    def raise_not_implemented(*args, **kwargs):
        raise NotImplementedError

    stat = raise_not_implemented
    lstat = raise_not_implemented
    open = raise_not_implemented
    listdir = raise_not_implemented
    scandir = raise_not_implemented
    chmod = raise_not_implemented
    lchmod = raise_not_implemented
    mkdir = raise_not_implemented
    unlink = raise_not_implemented
    link_to = raise_not_implemented
    rmdir = raise_not_implemented
    rename = raise_not_implemented
    replace = raise_not_implemented
    symlink = raise_not_implemented
    utime = raise_not_implemented
    readlink = raise_not_implemented


class AbstractPath(PosixPath, metaclass=abc.ABCMeta):
    """ An abstract class that overrides pathlib.PurePath and pathlib.Path
    methods to implement casting to the AbstractPath subclass

    """
    __slots__ = ()

    accessor_class = AbstractPathAccessor

    def __init__(self, *args, **kwargs) -> None:
        self._accessor = self.accessor_class(self)

    @abc.abstractmethod
    def new(self, path: _PT) -> AbstractPath:
        """ Return a new instance of AbstractPath

        It is useful for path initialized via object.__new__(cls) by PurePath (i.e. PurePath._from_parts).
        Inside this method we can initialize attributes that cannot be initialized because of PurePath initializing logic
        """
        raise NotImplementedError

    # below are overridden pathlib.PurePath methods
    def __truediv__(self, key) -> AbstractPath:
        return self.new(super().__truediv__(key))

    def __rtruediv__(self, key) -> AbstractPath:
        return self.new(super().__truediv__(key))

    def with_name(self, name) -> AbstractPath:
        return self.new(super().with_name(name))

    def with_suffix(self, suffix) -> AbstractPath:
        return self.new(super().with_suffix(suffix))

    def relative_to(self, *other) -> AbstractPath:
        return self.new(super().relative_to(other))

    def joinpath(self, *args) -> AbstractPath:
        return self.new(super().joinpath(args))

    @property
    def parent(self) -> AbstractPath:
        return self.new(super().parent)

    @property
    def parents(self) -> Sequence[AbstractPath]:
        return [self.new(parent) for parent in super().parents]

    # below are overridden pathlib.Path methods
    def _make_child_relpath(self, part) -> AbstractPath:
        return self.new(super()._make_child_relpath(part))

    def absolute(self) -> AbstractPath:
        return self.new(super().absolute())

    def resolve(self, strict=False) -> AbstractPath:
        return self.new(super().resolve(strict))

    def rename(self, target) -> AbstractPath:
        return self.new(super().rename(target))

    def replace(self, target) -> AbstractPath:
        return self.new(super().replace(target))

    def expanduser(self) -> AbstractPath:
        return self.new(super().expanduser())



import re
from setuptools import setup, find_packages

with open('pathliberty/_version.py', encoding='utf-8') as f:
    version = re.search(r'__version__ = \'([^\'"]+)\'', f.read()).group(1)

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='pathliberty',
    author = 'Egor Abramov',
    author_email = 'coreegor@gmail.com',
    version=version,
    description="pathlib extensions",
    packages=find_packages(),
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
    download_url=f'https://github.com/coreegor/pathliberty/archive/refs/tags/{version}.tar.gz',
    python_requires='>=3.8',
    install_requires=[
        'paramiko==2.9.2',
    ],
    project_urls={
        'Source': 'https://github.com/coreegor/pathliberty',
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ]

)



from __future__ import annotations
import contextlib
import fnmatch
from os import path
from pathlib import PosixPath
from time import time

import paramiko
from typing import Any, Callable, ContextManager, Dict, Iterable, Optional, Tuple, Union
from pathliberty import AbstractPathAccessor
from pathliberty import AbstractRemotePath
from pathliberty.ssh import SSHSession


class SSHAccessor(AbstractPathAccessor):
    """ Class for manipulating remote files over SSH """

    _ssh_clients: Dict[Tuple, paramiko.SSHClient] = {}

    #TODO: shared client object for same host, proxy
    # def __init__(self, path_obj: SSHPath) -> None:
    def __init__(self, path_obj: SSHPath) -> None:
        # self.host = path_obj.host
        # self.proxy = path_obj.proxy


        # if client := self._ssh_clients.get(self.client_key):
        #     self.ssh = client
        # else:
        #     self.ssh = get_ssh(
        #             self.host,
        #             jump_host=self.proxy,
        #         )
        #     self._ssh_clients[self.client_key] = self.ssh

        # self._sftp_open = False

        # self.client = path_obj.ssh.client
        self.sftp = path_obj.ssh.sftp


    # @property
    # def client_key(self):
    #     return (self.host, self.proxy)

    # @property
    # def sftp(self) -> paramiko.SFTPClient:
    #     if not self._sftp_open:
    #         try:
    #             self._sftp = self.ssh.open_sftp()
    #             self._sftp_open = True
    #         except Exception:
    #             self._sftp = None

    #     return self._sftp

    # def is_connected(self):
    #     return self.ssh.get_transport() is not None

    # def __del__(self):
    #     # if self.is_connected():
    #     #     if self.sftp:
    #     #         self.sftp.close()
    #     #     if self.ssh:
    #     #         self.ssh.close()
    #     print(f"\ndelete accessor {self}: {self._ssh_clients = }")
    #     # del self._ssh_clients[self.client_key]


    def stat(self, path):
        return self.sftp.stat(str(path))

    def lstat(self, path):
        return self.sftp.lstat(str(path))

    def open(self, mode='r', bufsize=-1, encoding=None, errors=None):
        if encoding:
            raise RuntimeError(
                f"{self.__class__.__name__} doesn't support an 'encoding' parameter. "
                f"The encoding of the file is assumed to be UTF-8."
            )
        if errors:
            raise RuntimeError(
                f"{self.__class__.__name__} doesn't support an 'errors' parameter."
            )
        return self.sftp.open(str(self), mode=mode, bufsize=bufsize)

    def listdir(self, path):
        return self.sftp.listdir(str(path))

    def scandir(self, path: SSHPath) -> ContextManager[Iterable[SSHPath]]:
        class _ScandirIterator:
            """ A temporary iterator to support the following pathlib constructions:
                with scandir(parent_path) as scandir_it:
                    entries = list(scandir_it)

            Objects of this class are intended to be used instead of os.DirEntry objects inside pathlib._Selector's
            """
            def __init__(this, path: SSHPath): this.path = path
            def __enter__(this): return this
            def __exit__(*args): pass
            def __iter__(this):
                new_path = this.path.new
                for p in self.listdir(path):
                    yield this.path / new_path(p)

        return _ScandirIterator(path)

    def chmod(self, mode):
        self.sftp.chmod(self(str), mode)

    def mkdir(self, path, mode=0o777):
        self.sftp.mkdir(str(path), mode=mode)

    def remove(self, path):
        self.sftp.remove(str(path))

    unlink = remove

    def rmdir(self, path: SSHPath):
        self.sftp.rmdir(str(path))

    def rename(self, src: SSHPath, dst):
        self.sftp.rename(str(src), dst)

    def rename(self, src: SSHPath, dst):
        self.unlink(dst)
        self.rename(src, dst)

    def symlink(self, target, source):
        self.sftp.symlink(str(source), target)

    def utime(self, path, times):
        self.sftp.utime(str(path), times)

    def readlink(self, path):
        return self.sftp.readlink(str(path))



class SSHPath(AbstractRemotePath):
    __slots__ = ('host', 'proxy', 'ssh')
    accessor_class = SSHAccessor

    def __init__(
        self,
        *pathsegments,
        host: Optional[str] = None,
        proxy: Optional[str] = None,
        ssh: Optional[SSHSession] = None,
    ):

        if host and ssh:
            raise RuntimeError("host and ssh arguments cannot be specified together")

        if not ssh and not host:
            raise RuntimeError("at least one of the host or ssh arguments must be specified")

        if not ssh:
            self.ssh = SSHSession(
                host,
                jump_host=proxy,
            )
        else:
            self.ssh = ssh

        self.host = self.ssh.host
        self.proxy = self.ssh.proxy

        super().__init__(pathsegments)

    def is_equal_to(self, other) -> bool:
        if (self.ssh.host, self.ssh.proxy) == (other.ssh.host, other.ssh.proxy):
            return True

    @property
    def sftp(self) -> paramiko.SFTPClient:
        return self.ssh.sftp

    def new(self, path: Union[str, AbstractRemotePath]) -> SSHPath:
        return SSHPath(path, ssh=self.ssh)

    # TODO: check
    def cwd(self):
        """ Return a new path pointing to the "current working directory"
        for this SFTP session
        """
        return self.new(self.sftp.getcwd())

    # TODO: check
    def home(self):
        """ Return a new path pointing to the user's home directory on remote
        """
        return self.new(self.ssh.home)

    def rmdir(self, recursive: bool = False):
        """ Remove this directory.

        Optionally do it recursively.
        """
        if recursive:
            for p in self.iterdir():
                if p.is_dir():
                    p.rmdir(recursive=True)
                else:
                    p.unlink()
            # if directory is empty
            super().rmdir()
        else:
            super().rmdir()


    def open(self, mode='r', buffering=-1) -> paramiko.SFTPFile:
        return self.sftp.open(str(self), mode=mode, bufsize=buffering)

    def touch(self, exist_ok=True):
        try:
            with self.open(mode='x'):
                pass
        except OSError:
            if exist_ok:
                self.sftp.utime(str(self), None)
            else:
                raise FileExistsError(str(self))

    def rename(self, target):
        return self.new(super().rename(target))

    def replace(self, target):
        return self.new(super().replace(target))

    # def glob(self, pattern: str):
    #     return fnmatch.filter(self.sftp.listdir(str(self)), pattern)

    # def recursive_iterdir(self, l = None):
    #     if not l: l = []
    #     if self.is_dir():
    #         for p in self.iterdir():
    #             if p.is_dir():
    #                 l.extend(p.recursive_iterdir(l))
    #             else:
    #                 l.append(p)
    #     else:
    #         l.append(self)
    #     return l
    # def rglob(self, pattern: str):
    #     l = self.recursive_iterdir()
    #     return fnmatch.filter(map(str, l), pattern)



import paramiko
import getpass
from typing import Optional
from pathliberty import ssh_config

def get_ssh_client(
    dest_host: str,
    *,
    username: str = getpass.getuser(),
    password: Optional[str] = None,
    dest_port: Optional[int] = None,
) -> paramiko.SSHClient:
    """ Get a high-level representation of a open session with an SSH server.

    This object can be used as contextmanager::

        with get_ssh_client('hostname') as client:
            stdin, stdout, stderr = client.exec_command('ls -al')
            print(stdout.read().decode("utf-8"))

    Args:
        dest_host (str): target host
        username (str): the username to authenticate as
        password (str): Used for password authentication
        dest_port (int, optional): target port. If None, it will be a ssh_config.DEFAULT_PORT.

    Returns:
        paramiko.SSHClient: Object of  :class:`~paramiko.SSHClient`
    """
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    client.connect(
        hostname=dest_host,
        username=username,
        password=password,
        port=dest_port or ssh_config.DEFAULT_PORT,
        timeout=ssh_config.DEFAULT_PORT
    )
    return client

def get_ssh_client_through_proxy(
    dest_host: str,
    *,
    jump_host: str,
    username: str = getpass.getuser(),
    password: Optional[str] = None,
    dest_port: Optional[int] = None,
    jump_port: Optional[int] = None,
) -> paramiko.SSHClient:
    """ Get a high-level representation of a open session with an SSH server via jump-host

    This object can be used as contextmanager.

    Args:
        dest_host (str): target host
        jump_host (str): jump host used as transport
        username (str): the username to authenticate as
        password (str): Used for password authentication
        dest_port (int, optional): target port. If None, it will be a ssh_config.DEFAULT_PORT.
        jump_port (int, optional): jump port. If None, it will be a ssh_config.DEFAULT_PORT.

    Returns:
        paramiko.SSHClient: Object of :class:`~paramiko.SSHClient`
    """
    dest_port = dest_port or ssh_config.DEFAULT_PORT
    jump_port = jump_port or ssh_config.DEFAULT_PORT
    
    jclient = paramiko.SSHClient()
    jclient.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    jclient.connect(hostname=jump_host, username=username, password=password, port=jump_port, timeout=ssh_config.DEFAULT_CONNECTION_TIMEOUT)

    jtransport = jclient.get_transport()
    dest_addr = (dest_host, dest_port)
    local_addr = (jump_host, jump_port)
    jchannel = jtransport.open_channel("direct-tcpip", dest_addr, local_addr, timeout=ssh_config.DEFAULT_CONNECTION_TIMEOUT)

    target = paramiko.SSHClient()
    target.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    target.connect(
        hostname=jump_host,
        username=username,
        port=jump_port,
        sock=jchannel,
        password=password,
        timeout=ssh_config.DEFAULT_CONNECTION_TIMEOUT
    )

    return target

def get_ssh(
    dest_host: str,
    *,
    username: str = getpass.getuser(),
    password: Optional[str] = None,
    dest_port: Optional[int] = None,
    jump_host: Optional[str] = None,
    jump_port: Optional[int] = None,
) -> paramiko.SSHClient:

    try:
        if jump_host:
            client = get_ssh_client_through_proxy(
                dest_host=dest_host,
                dest_port=dest_port,
                username=username,
                password=password,
                jump_host=jump_host,
                jump_port=jump_port,
            )
        else:
            client = get_ssh_client(
                dest_host=dest_host,
                dest_port=dest_port,
                username=username,
                password=password,
            )
    except paramiko.AuthenticationException as e:
        base_msg = e.args[0]
        if not password:
            raise RuntimeError(f"{base_msg} Try to use password or check your keys")
        else:
            raise RuntimeError(f"{base_msg} Check your password or try to use keys")

    else:
        return client

class SSHSession:
    def __init__(
        self,
        dest_host: str,
        *,
        username: str = getpass.getuser(),
        password: Optional[str] = None,
        dest_port: Optional[int] = None,
        jump_host: Optional[str] = None,
        jump_port: Optional[int] = None,
    ):

        self.client = get_ssh(
            dest_host,
            username=username,
            password=password,
            dest_port=dest_port,
            jump_host=jump_host,
            jump_port=jump_port,
        )

        self.host = dest_host
        self.proxy = jump_host

        self._sftp_open = False

    @property
    def sftp(self) -> paramiko.SFTPClient:
        if not self._sftp_open:
            self._sftp = self.client.open_sftp()
            self._sftp_open = True

        return self._sftp

    def __del__(self):
        self.client.close()

    def exec(self, command: str) -> str:
        _, stdout, stderr = self.client.exec_command(
            command=command,
            timeout=ssh_config.DEFAULT_CONNECTION_TIMEOUT
        )

        status = stdout.channel.recv_exit_status()

        if not status:
            return stdout.read().decode().strip()
        else:
            raise RuntimeError(stderr.read().decode())

    @property
    def home(self):
        return self.exec('echo ~').strip()
